
import { URL } from 'url';
import { request as https_request } from 'https';
import { request as http_request, OutgoingHttpHeaders, IncomingHttpHeaders } from 'http';
import { ResponseCacheConfig, create_response_cache, validate_response_cache_conf } from './response-cache';
import { noop_logger } from './noop-logger';
import type { HttpURL, Logger } from '@js/types';

export { ResponseCacheConfig, validate_response_cache_conf } from './response-cache';

let next_id = 1;
const timeout = 5000;

export interface HttpResult {
	url: URL;
	status: number;
	headers: IncomingHttpHeaders;
	content: string;
}

export interface HttpClientConfig {
	logger?: Logger;
	https_only?: boolean;
	cache?: ResponseCacheConfig;
}

export function validate_http_conf(conf: unknown) : asserts conf is HttpClientConfig {
	if (typeof conf !== 'object' || ! conf) {
		throw new Error('Configuration must be an object');
	}

	if ('logger' in conf) {
		// todo: validate
	}

	if ('https_only' in conf) {
		if (typeof conf.https_only !== 'boolean') {
			throw new Error('"https_only" must be a boolean if provided');
		}
	}

	if ('cache' in conf) {
		validate_response_cache_conf(conf.cache);
	}
}

export type HttpClient = ReturnType<typeof create_http_client>;

export function create_http_client(conf: HttpClientConfig) {
	const { logger = noop_logger } = conf;
	const response_cache = create_response_cache(conf.cache ?? { }, logger);

	return {	
		req: http_req,	
		get(url_str: HttpURL, headers?: OutgoingHttpHeaders, skip_cache = false) {
			return http_req('GET', url_str, headers, null, skip_cache);
		},
		head(url_str: HttpURL, headers?: OutgoingHttpHeaders, skip_cache = false) {
			return http_req('HEAD', url_str, headers, null, skip_cache);
		},
		post(url_str: HttpURL, headers?: OutgoingHttpHeaders, body?: string, skip_cache = false) {
			return http_req('POST', url_str, headers, body, skip_cache);
		},
		put(url_str: HttpURL, headers?: OutgoingHttpHeaders, body?: string, skip_cache = false) {
			return http_req('PUT', url_str, headers, body, skip_cache);
		},
		patch(url_str: HttpURL, headers?: OutgoingHttpHeaders, body?: string, skip_cache = false) {
			return http_req('PATCH', url_str, headers, body, skip_cache);
		},
		delete(url_str: HttpURL, headers?: OutgoingHttpHeaders, body?: string, skip_cache = false) {
			return http_req('DELETE', url_str, headers, body, skip_cache);
		},
	};

	async function http_req(method: string, url_str: HttpURL, headers?: OutgoingHttpHeaders, body?: string, skip_cache = false) {
		const url = new URL(url_str);
		const make_request
			= url.protocol === 'https:' ? https_request
			: url.protocol === 'http:' ? http_request
			: null;
	
		if (! make_request) {
			throw new Error('illegal protocol');
		}
	
		if (url.protocol !== 'https:' && conf.https_only) {
			throw new Error('non-secure http requests not allowed (by configuration)');
		}
	
		if (url.username || url.password) {
			throw new Error('urls containing user credentials not allowed');
		}

		const req_headers = headers ? structuredClone(headers) : { };
	
		const result: HttpResult = {
			url,
			status: null,
			content: null,
			headers: null,
		};
	
		const path = url.pathname + (url.search || '');
		const port = url.port ? parseInt(url.port, 10) : (url.protocol === 'https:' ? 443 : 80);
		const log = logger.child({
			req_id: next_id++,
			method: method,
			protocol: url.protocol,
			hostname: url.hostname,
			port: port,
			path: path,
		});

		const cached = skip_cache
			? null
			: response_cache.check_for_cached_response(method, url_str, req_headers);

		if (cached) {
			const revalidation = response_cache.determine_revalidation_needed(cached);

			if (! revalidation.should_revalidate) {
				log.info('serving request from cache (no revalidation required)');
				result.status = cached.status;
				result.content = cached.body;
				result.headers = structuredClone(cached.res_headers);
				return Promise.resolve(result);
			}

			// if (! revalidation.must_revalidate) {
			// 	// todo: lazy revalidation option
			// }

			log.info('cached response found, but requires revalidation');
			Object.assign(req_headers, revalidation.headers);
		}
	
		log.info('starting outbound http request');
	
		return new Promise<HttpResult>((resolve, reject) => {
			const start = Date.now();
			const req = make_request({
				method: method,
				protocol: url.protocol,
				hostname: url.hostname,
				port: port,
				path: path,
				headers: req_headers
			}, (res) => {
				// result.res = res;
				result.status = res.statusCode;

				log.info({
					status: res.statusCode,
					duration: Date.now() - start,
				}, `headers received`);

				switch (result.status) {
					case 304:
						log.info({ status: cached.status }, 're-using cached response (not modified)');
						result.status = cached.status;
						result.content = cached.body;
						result.headers = structuredClone(cached.res_headers);
						response_cache.freshen_cache(cached, res.headers);
						resolve(result);
						break;

					case 500:
					case 502:
					case 503:
					case 504:
						if (cached?.cache_info?.cache_control?.stale_if_error) {
							log.info({ status: cached.status }, 're-using cached response (response returned an error status, and `stale-if-error` directive was provided)');
							result.status = cached.status;
							result.content = cached.body;
							result.headers = structuredClone(cached.res_headers);
							resolve(result);
							break;
						}
						// nobreak

					default: {
						const start = Date.now();
						let data = '';
			
						res.on('data', (chunk) => {
							data += chunk;
						});
			
						res.on('end', () => {
							log.info({ duration: Date.now() - start }, `response finished`);
							result.content = data;
							result.headers = res.headers;
							response_cache.add_response_to_cache(method, url_str, req_headers, result);
							resolve(result);
						});
					};
				}
			});
	
			req.setTimeout(timeout, () => {
				log.error('request timeout');
				req.destroy();
			});
		
			req.on('error', (error) => {
				log.error('request error ' + error.stack);
				reject(error);
			});
	
			if (body) {
				req.write(body);
			}
	
			req.end();
		});
	}
}
