
import type { IncomingHttpHeaders } from 'http';

export interface ParsedCacheHeaders {
	age?: number;
	etag?: string;
	vary?: string[];
	date?: Date;
	cache_control?: CacheControl;
	expires?: Date;
	last_modified?: Date;
}

export interface CacheControl {
	public?: boolean;
	private?: boolean;
	no_store?: boolean;
	no_cache?: boolean;
	max_age?: number;
	s_max_age?: number;
	must_revalidate?: boolean;
	proxy_revalidate?: boolean;
	immutable?: boolean;
	stale_while_revalidate?: boolean;
	stale_if_error?: boolean;
	must_understand?: boolean;
}

export function parse_cache_headers(headers: IncomingHttpHeaders) {
	const result: ParsedCacheHeaders = { };

	if (headers['age']) {
		result.age = parseInt(headers['age'], 10);
	}

	if (headers['etag']) {
		result.etag = headers['etag'];
	}

	if (headers['vary']) {
		result.vary = headers['vary'].split(',').map((str) => str.trim().toLowerCase());
	}

	if (headers['date']) {
		result.date = new Date(Date.parse(headers['date']));
	}

	if (headers['cache-control']) {
		result.cache_control = { };

		for (let directive of headers['cache-control'].split(',')) {
			directive = directive.trim();

			switch (directive) {
				case 'public':
					result.cache_control.public = true;
					break;
				case 'private':
					result.cache_control.private = true;
					break;
				case 'no-store':
					result.cache_control.no_store = true;
					break;
				case 'no-cache':
					result.cache_control.no_cache = true;
					break;
				case 'must-revalidate':
					result.cache_control.must_revalidate = true;
					break;
				case 'proxy-revalidate':
					result.cache_control.proxy_revalidate = true;
					break;
				case 'immutable':
					result.cache_control.immutable = true;
					break;
				case 'stale-while-revalidate':
					result.cache_control.stale_while_revalidate = true;
					break;
				case 'stale-if-error':
					result.cache_control.stale_if_error = true;
					break;
				case 'must-understand':
					result.cache_control.must_understand = true;
					break;
				default:
					if (directive.startsWith('max-age=')) {
						result.cache_control.max_age = parseInt(directive.slice(8), 10);
						break;
					}

					if (directive.startsWith('s-maxage=')) {
						result.cache_control.s_max_age = parseInt(directive.slice(10), 10);
						break;
					}

					// todo: log something here about unknown directive
			}
		}
	}

	if (headers['expires']) {
		result.expires = new Date(Date.parse(headers['expires']));
	}

	if (headers['last-modified']) {
		result.last_modified = new Date(Date.parse(headers['last-modified']));
	}

	return result;
}
