
import { parse_cache_headers, ParsedCacheHeaders } from './parse-cache-headers';
import type { HttpResult } from './index';
import type { HttpURL, Logger } from '@js/types';
import type { OutgoingHttpHeaders, IncomingHttpHeaders } from 'http';

export interface ResponseCacheConfig {
	cacheable_methods?: string[];
	cacheable_status_codes?: number[];
	cache_invalidating_methods?: string[];
	// todo: support for configurable limits:
	max_records?: number;  // max number of cached responses
	max_length?: number;  // max content-length for one response
	evict_after?: number;  // period of time unused after which a response should be evicted
}

export interface CachedResponse {
	cache_key: string;
	cached_time: Date;
	is_stale: boolean;
	url: HttpURL;
	method: string;
	status: number;
	body: string;
	res_headers: IncomingHttpHeaders;
	cache_info: ParsedCacheHeaders;
	expire_timer: NodeJS.Timeout;
}

export function validate_response_cache_conf(conf: unknown) : asserts conf is ResponseCacheConfig {
	// todo: validate config
}

export type ResponseCache = ReturnType<typeof create_response_cache>;

export function create_response_cache(conf: ResponseCacheConfig, logger: Logger) {
	type URLCache = Record<string, CachedResponse>;
	const cached_responses: Record<HttpURL, URLCache> = Object.create(null);

	const cacheable_methods = new Set(conf.cacheable_methods ?? [ 'GET', 'HEAD' ]);
	const cache_invalidating_methods = new Set(conf.cache_invalidating_methods ?? [ 'POST', 'PUT', 'PATCH', 'DELETE' ]);
	const cacheable_status_codes = new Set(conf.cacheable_status_codes ?? [
		200, 203, 204, 206, 300, 301, 308, 404, 405, 410, 414, 501
	]);

	return {
		/**
		 * Should be called before making a new request to check if a cached response already exists
		 */
		check_for_cached_response(method: string, url_str: HttpURL, headers?: OutgoingHttpHeaders) : Readonly<CachedResponse> {
			if (! cached_responses[url_str]) {
				return;
			}

			for (const [cache_key, cached] of Object.entries(cached_responses[url_str])) {
				if (cache_key === header_cache_key(method, cached.cache_info.vary, headers)) {
					return cached;
				}
			}
		},

		/**
		 * Should be called when a cached response is found, to determine what if any revalidation is
		 * required before re-using the response
		 */
		determine_revalidation_needed(cached: CachedResponse) {
			const { cache_control, etag, date, last_modified } = cached.cache_info;

			const must_revalidate = Boolean(cache_control?.must_revalidate || cache_control?.no_cache);
			const should_revalidate = must_revalidate || cached.is_stale;
			const headers: OutgoingHttpHeaders = { };

			if (etag) {
				headers['if-none-match'] = etag;
			}

			if (last_modified || date) {
				headers['if-modified-since'] = last_modified?.toUTCString() ?? date?.toUTCString();
			}

			else {
				headers['if-modified-since'] = cached.cached_time.toUTCString();
			}

			return {
				must_revalidate,
				should_revalidate,
				headers,
			};
		},

		/**
		 * Should be called when receiving a non-`304 Not Modified` response from a request, either
		 * as a novel request, or during a revalidation request
		 */
		add_response_to_cache(method: string, url_str: HttpURL, headers: OutgoingHttpHeaders, result: HttpResult) {
			if (! cacheable_methods.has(method) || ! cacheable_status_codes.has(result.status)) {
				logger.info({ method, status: result.status }, 'method/status not cacheable');
				return;
			}

			const cache_info = parse_cache_headers(result.headers);

			if (headers.authorization) {
				// todo: caching authenticated requests
				// see <https://www.rfc-editor.org/rfc/rfc9111#section-3.5>
				return;
			}

			// todo: if `vary` has changed, invalidate all stored responses
			// todo: source link for ^

			if (cache_info.vary.includes('*')) {
				logger.info(cache_info, 'response contains "Vary: *" header that prevents caching');
				return;
			}

			if (cache_info.cache_control?.no_store || cache_info.cache_control?.private) {
				logger.info(cache_info, 'response contains "Cache-Control" directives that prevent caching');
				return;
			}
			
			const cache_key = header_cache_key(method, cache_info.vary, headers);
			unset_expire_timer(url_str, cache_key);

			const cached: CachedResponse = {
				cache_key,
				cached_time: new Date(Date.now()),
				is_stale: false,
				url: url_str,
				method,
				// req_headers: structuredClone(headers),
				status: result.status,
				body: result.content,
				res_headers: structuredClone(result.headers),
				cache_info,
				expire_timer: null,
			};

			logger.info(cache_info, 'caching response');
			store_to_cache(cached, cache_info);
		},

		/**
		 * Should be called when receiving a `304 Not Modified` response during a revalidation request
		 * to update the cache with more recent freshness info
		 */
		freshen_cache(cached: CachedResponse, headers: IncomingHttpHeaders) {
			const cache_info = parse_cache_headers(headers);
			unset_expire_timer(cached.url, cached.cache_key);
			store_to_cache(cached, cache_info);
		},

		/**
		 * 
		 */
		invalidate_cache_for_update(method: string, url_str: HttpURL, status: number, headers: IncomingHttpHeaders) {
			if (! cache_invalidating_methods.has(method)) {
				return;
			}

			if (status < 200 || status >= 400) {
				return;
			}
		},
	};

	function store_to_cache(cached: CachedResponse, cache_info: ParsedCacheHeaders) {
		if (! cached_responses[cached.url]) {
			cached_responses[cached.url] = Object.create(null);
		}

		cached_responses[cached.url][cached.cache_key] = cached;
		cached.is_stale = false;

		if (cache_info.cache_control?.max_age) {
			let ttl_ms = cache_info.cache_control.max_age * 1000;

			if (cache_info.age) {
				ttl_ms -= cache_info.age * 1000;
			}

			if (ttl_ms <= 0) {
				cached.is_stale = true;
				return;
			}

			cached.expire_timer = set_expire_timer(cached.url, cached.cache_key, ttl_ms);
		}

		else if (cache_info.expires) {
			const ttl_ms = cache_info.expires.getTime() - Date.now();

			if (ttl_ms <= 0) {
				cached.is_stale = true;
				return;
			}

			cached.expire_timer = set_expire_timer(cached.url, cached.cache_key, ttl_ms);
		}

		// If no cache-control or expires header is given, default to the
		// recommended 10% of time since last-modified
		else if (cache_info.last_modified) {
			const last_modified_ago_ms = Date.now() - cache_info.last_modified.getTime();

			if (last_modified_ago_ms <= 0) {
				cached.is_stale = true;
				return;
			}

			const ttl_ms = (last_modified_ago_ms / 10) | 0;
			cached.expire_timer = set_expire_timer(cached.url, cached.cache_key, ttl_ms);
		}
	}

	function unset_expire_timer(url_str: HttpURL, cache_key: string) {
		const cached = cached_responses[url_str]?.[cache_key];

		if (cached?.expire_timer) {
			clearTimeout(cached.expire_timer);
		}
	}

	function set_expire_timer(url_str: HttpURL, cache_key: string, ttl_ms: number) {
		const timer = setTimeout(expire, ttl_ms);
		timer.unref();
		return timer;

		function expire() {
			if (cached_responses[url_str]?.[cache_key]) {
				cached_responses[url_str][cache_key].is_stale = true;
			}
		};
	}
}

function header_cache_key(method: string, vary: string[], headers: OutgoingHttpHeaders) {
	const parts: string[] = [ method ];

	if (vary?.length) {
		for (const header of vary) {
			if (Array.isArray(headers[header])) {
				const values = (headers[header] as string[]).sort();
				for (const value of values) {
					parts.push(`${header}: ${value}`);
				}
			}

			else if (headers[header]) {
				parts.push(`${header}: ${headers[header]}`);
			}
		}
	}

	return parts.join('\0');
}
